// Copyright (C) Thorsten Thormaehlen

#include "pch.h"

#include <string>
#include <iostream> 
#include <fstream> 

using namespace std;

#include "BVHierarchy.h"

#define RTEXPF_ERROR(msg) { cout << "ERROR: "   << msg << " (" << __FILE__ << ":"<< __LINE__ << ")" << endl; }  
#define RTEXPF_WARN(msg) { cout << "WARNING: " << msg << " (" << __FILE__ << ":"<< __LINE__ << ")" << endl; }  
#define RTEXPF_DEBUG(msg) { cout << "DEBUG: "   << msg << " (" << __FILE__ << ":"<< __LINE__ << ")" << endl; } 
#define RTEXPF_INFO(msg) { cout << "INFO: "   << msg << " (" << __FILE__ << ":"<< __LINE__ << ")" << endl;}


// some parameters
const float tbb = 1.0; // time required to interest the bounding box
const float tTri = 1.0; // time required to interest another object (e.g. a triangle)
const bool debugOn = false;

// some static helper functions for sorting
static bool SceneObjectSortX(const Intersectable* d1, const Intersectable* d2) {
  float x1 = d1->getBoundingBox().center().x;
  float x2 = d2->getBoundingBox().center().x;
  return x1 < x2;
}
static bool SceneObjectSortY(const Intersectable* d1, const Intersectable* d2) {
  float y1 = d1->getBoundingBox().center().y;
  float y2 = d2->getBoundingBox().center().y;
  return y1 < y2;
}
static bool SceneObjectSortZ(const Intersectable* d1, const Intersectable* d2) {
  float z1 = d1->getBoundingBox().center().z;
  float z2 = d2->getBoundingBox().center().z;
  return z1 < z2;
}


bool BVHierarchyNode::depthSearch(const Ray& ray, RayHit &hit, float &closest) const
{
  bool ret = false;
  if(box.intersect(ray)) {
    if(leftNode == NULL && rightNode == NULL) { // node is tree leaf
      for(unsigned k=0; k < enclosedObjects.size(); k++) {
        RayHit candidateHit;
        bool b = enclosedObjects[k]->intersect(ray, candidateHit);
        if(b) {
          float dist = (ray.origin-candidateHit.hitPoint).length();
          if(dist < closest) {
            hit.hitPoint = candidateHit.hitPoint;
            hit.ray = ray;
            closest = dist;
            ret = true;
          } 
        }
      }
    } else { // node is inner node
      bool retLeft, retRight; 
      if(leftNode != NULL) {
        retLeft = leftNode->depthSearch(ray, hit, closest);
      }
      if(rightNode != NULL) {
        retRight = rightNode->depthSearch(ray, hit, closest);
      }
      ret = (retLeft || retRight);
    }
  }
  return ret;
}

bool BVHierarchyNode::getClosestPointInBoxDepthSearch(Vec3f& p, BoundingBox& searchBox, float& closestDistance, Vec3f& closestPoint) const {
	bool ret = false;
	if (box.intersect(searchBox)) {
		if (leftNode == NULL && rightNode == NULL) { // node is tree leaf
			for (unsigned k = 0; k < enclosedObjects.size(); k++) {
				Vec3f candidatePoint = enclosedObjects[k]->getClosestPoint(p);
				float candidateDistance = (candidatePoint - p).length();
				if (candidateDistance < closestDistance) {
					closestDistance = candidateDistance;
					closestPoint = candidatePoint;
					ret = true;
				}
			}
		}
		else { // node is inner node
			bool retLeft, retRight;
			Vec3f c;
			if (leftNode != NULL) {
				retLeft = leftNode->getClosestPointInBoxDepthSearch(p, searchBox, closestDistance, closestPoint);
			}
			if (rightNode != NULL) {

				retRight = rightNode->getClosestPointInBoxDepthSearch(p, searchBox, closestDistance, closestPoint);
			}
			ret = (retLeft || retRight);
		}
	}
	return ret;
}


Triangle* BVHierarchyNode::getAdjacentEdgeDepthSearch(Triangle* excludeTriangle, BoundingBox& searchBox, Vec3f& edgeA, Vec3f& edgeB, float& closestDistance) const {
	Triangle* ret = NULL;
	if (box.intersect(searchBox)) {
		if (leftNode == NULL && rightNode == NULL) { // node is tree leaf
			for (unsigned k = 0; k < enclosedObjects.size(); k++) {
				if (typeid(*enclosedObjects[k]) != typeid(Triangle))
					continue;
				Triangle* t = (Triangle*)enclosedObjects[k];
				if (t == excludeTriangle)
					continue;
				float da = min((edgeA - t->a).sqrLength(), min((edgeA - t->b).sqrLength(), (edgeA - t->c).sqrLength()));
				float db = min((edgeB - t->a).sqrLength(), min((edgeB - t->b).sqrLength(), (edgeB - t->c).sqrLength()));
				float d = max(da, db);

				if (d < closestDistance) {
					closestDistance = d;
					ret = t;
				}

			}
		}
		else { // node is inner node
			Triangle* retLeft = NULL;
			Triangle* retRight = NULL;
			if (leftNode != NULL) {
				retLeft = leftNode->getAdjacentEdgeDepthSearch(excludeTriangle, searchBox, edgeA, edgeB,closestDistance);
			}
			if (rightNode != NULL) {
				retRight = rightNode->getAdjacentEdgeDepthSearch(excludeTriangle, searchBox, edgeA, edgeB, closestDistance);
			}
			if (retRight != NULL)
				ret = retRight;
			else if (retLeft != NULL)
				ret = retLeft;
		}
	}
	return ret;
}

//trys to find the adjacent edge resultA-resultB and the parent Triangle "resultParentTriangle" for the given edge a-b witht the given parent "parentTriangle" and returns true if such an edge exists, false otherwise.
bool BVHierarchyNode::getAdjacentEdge(Vec3f& a, Vec3f& b, Triangle* parentTriangle, Triangle*& resultParentTriangle) const {
	BoundingBox bb;
	bb.min.x = min(a.x-0.01f, b.x - 0.01f); bb.min.y = min(a.y - 0.01f, b.y - 0.01f); bb.min.z = min(a.z - 0.01f, b.z - 0.01f);
	bb.max.x = max(a.x + 0.01f, b.x + 0.01f); bb.max.y = max(a.y + 0.01f, b.y + 0.01f); bb.max.z = max(a.z + 0.01f, b.z + 0.01f);
	float distance = FLT_MAX;
	Triangle* t = getAdjacentEdgeDepthSearch(parentTriangle, bb, a, b, distance);
	Vec3f resultA, resultB;

	if(t==NULL)
		return false;

	resultParentTriangle = t;

	//get closest point to a
	float d = (a - t->a).sqrLength();
	float minDA = d;
	resultA = t->a;

	d = (a - t->b).sqrLength();
	if (d < minDA) {
		minDA = d;
		resultA = t->b;
	}

	d = (a - t->c).sqrLength();
	if (d < minDA) {
		minDA = d;
		resultA = t->c;
	}

	//get closest point to b
	d = (b - t->a).sqrLength();
	float minDB = d;
	resultA = t->a;

	d = (b - t->b).sqrLength();
	if (d < minDB) {
		minDB = d;
		resultA = t->b;
	}

	d = (b - t->c).sqrLength();
	if (d < minDB) {
		minDB = d;
		resultA = t->c;
	}

	if (max(minDA, minDB) > 0.001f)
		return false;

	return true;
}

Vec3f BVHierarchyNode::getClosestPoint(const Vec3f& p) const {
	return Vec3f(0,0,0);
}

void BVHierarchyNode::build(std::vector<SceneObject*> &objects) 
{
  //RTEXPF_INFO("Building Bounding Volume Hierarchy for " << objects.size() << " objects ...");

  // split objects in those that are interectable and others
  std::vector<Intersectable*> interectableObjects;
  std::vector<SceneObject*> others;
  for(unsigned i=0; i < objects.size(); i++) {
    Intersectable* intersectable = dynamic_cast<Intersectable*>(objects[i]);
    if(intersectable) {
      interectableObjects.push_back(intersectable);
    }else{
      others.push_back(objects[i]);
    }
  }

  // remove interectable objects from the given objects
  objects.clear();
  for(unsigned i=0; i < others.size(); i++) {
    objects.push_back(others[i]);
  }

  partitionSweep(interectableObjects, 0);
  //RTEXPF_INFO("Done ...");
}

Vec3f BVHierarchyNode::getCenter() const {
	return box.center();
}

float BVHierarchyNode::getSurfaceArea() const {
	return box.surfaceArea();
}

BoundingBox BVHierarchyNode::getUnionOfBoundingBoxes(std::vector<Intersectable*> &objects) 
{
  BoundingBox aabb;
  if(objects.size() == 0) return aabb;
  aabb = objects[0]->getBoundingBox();
  for(unsigned i=1; i < objects.size(); i++) {
    aabb = aabb.getUnion(objects[i]->getBoundingBox());
  }
  return aabb;
}

float BVHierarchyNode::surfaceAreaHeuristic(unsigned sizeLeft, float leftArea, unsigned sizeRight, float rightArea) 
{
      if(sizeLeft == 0 || sizeRight == 0) return 1e16f;
      const float totalArea = box.surfaceArea();
      return (2.0f * tbb + leftArea / totalArea * sizeLeft * tTri + rightArea / totalArea * sizeRight * tTri);
}

void BVHierarchyNode::partitionSweep(std::vector<Intersectable*> &objects, unsigned depth) 
{
  // build BVH according to  
  // Wald et al.: "Ray Tracing Deformable Scenes using Dynamic Bounding Volume Hierarchies", ToG 2007
  // Algorithm 1. Centroid-based SAH partitioning

  // compute union of bounding boxes of all interestable objects
  box = getUnionOfBoundingBoxes(objects);

  // sort objects using centroid of boxes in current axis
  std::vector <Intersectable*> sorted[3];
  // sort in x direction
  sorted[0].resize(objects.size());
  std::copy(objects.begin(), objects.end(), sorted[0].begin());
  std::sort(sorted[0].begin(), sorted[0].end(), SceneObjectSortX);

  // sort in y direction
  sorted[1].resize(objects.size());
  std::copy(objects.begin(), objects.end(), sorted[1].begin());
  std::sort(sorted[1].begin(), sorted[1].end(), SceneObjectSortY);

  // sort in z direction
  sorted[2].resize(objects.size());
  std::copy(objects.begin(), objects.end(), sorted[2].begin());
  std::sort(sorted[2].begin(), sorted[2].end(), SceneObjectSortZ);

  // create two sets
  std::vector<Intersectable*> left;
  std::vector<float> leftArea;
  std::vector<Intersectable*> right;
  std::vector<float> rightArea;

  float bestCost = tTri * objects.size(); // cost of making a leaf
  int bestAxis = -1;
  int bestEvent = -1;

  for(unsigned axis=0; axis < 3; axis++) { // trying to split along x-, y-, and z-axis
    left.clear();
    leftArea.clear();
    right.clear();
    rightArea.clear();

	BoundingBox aabb;

    // sweep from left
    for(unsigned i=0; i < sorted[axis].size(); i++) {
		float area = 1e32f;
		if (i > 0) {
			aabb = aabb.getUnion(sorted[axis][i]->getBoundingBox());
			area = aabb.surfaceArea();
		}
		else {
			aabb = sorted[axis][i]->getBoundingBox();
		}
		left.push_back(sorted[axis][i]);
		leftArea.push_back(area);
    }

    if(true) {
      // sweep from right
	  if(sorted[axis].size()>0)
		aabb = sorted[axis][sorted[axis].size() - 1]->getBoundingBox();
      for(int i=sorted[axis].size()-1; i >= 1; i--) {
		right.push_back(sorted[axis][i]);
		aabb = aabb.getUnion(sorted[axis][i]->getBoundingBox());
		float area = aabb.surfaceArea();
		float thisCost = surfaceAreaHeuristic(i, leftArea[i], right.size(), area);
		rightArea.push_back(area);
        if(thisCost < bestCost) {
          bestCost = thisCost;
          bestEvent = i;
          bestAxis = axis;
        }
      }
    }
  }

  if(bestAxis == -1) {
    // found no partition better than leaf
    
    enclosedObjects.clear();
    enclosedObjects.resize(objects.size());
    std::copy(objects.begin(), objects.end(), enclosedObjects.begin());
    if(debugOn) {
      //RTEXPF_DEBUG("depth=" << depth << " creating leaf node with bounding area " << box.surfaceArea() << " and " << objects.size() << " enclosed objects");
      //debug(depth);
    }
  } else {
    
    // make inner nodes with axis bestAxis;
    if(debugOn) {
       //RTEXPF_DEBUG("depth=" << depth << " creating inner tree node with bounding area " << box.surfaceArea());
       //debug(depth);
    }
   

    leftNode = new BVHierarchyNode();
    rightNode = new BVHierarchyNode();
    left.resize(bestEvent);
    std::copy(sorted[bestAxis].begin(), sorted[bestAxis].begin()+bestEvent, left.begin());
    right.resize(sorted[bestAxis].size()-bestEvent);
    std::copy(sorted[bestAxis].begin()+bestEvent, sorted[bestAxis].end(), right.begin());
    unsigned newDepth = depth+1;
    leftNode->partitionSweep(left, newDepth);
    rightNode->partitionSweep(right, newDepth);
  }
  
}

void BVHierarchyNode::debug(unsigned depth) const
{
  // export BVH to max script for debugging purpose
  ofstream myfile;
  if(depth == 0) {
    myfile.open ("scene.ms");
    myfile << "a=0;" << endl;
    myfile << "file =\"out\""<< endl;
    myfile << "ext =\".png\""<< endl;

  }else{
    // append to file
    myfile.open ("scene.ms", ios::app);
  }
  myfile << "m = standard shaderType:1" << endl;
  if(depth % 3 == 0) {
    myfile << "m.diffuse = (color 255 0 0)" << endl;
  }
  if(depth % 3 == 1) {
    myfile << "m.diffuse = (color 0 255 0)" << endl;
  }
  if(depth % 3 == 2) {
    myfile << "m.diffuse = (color 0 0 255)" << endl;
  }
  float opa = 10.0f*(depth+1);
  if(opa > 60.0) opa = 60.0;
  myfile << "m.opacity = "<< opa << endl;
  Vec3f dim = box.max - box.min;
  Vec3f c = box.center();
  c.z = box.min.z;
  myfile << "box lengthsegs:1 widthsegs:1 heightsegs:1 length:" << dim.y << " width:" << dim.x << " height:" << dim.z << " ";
  myfile << "mapcoords:on pos:[" << c.x << "," << c.y << "," << c.z << "] material:m"<< endl;
  for(unsigned k=0; k < enclosedObjects.size(); k++) {
    Triangle* tri = dynamic_cast<Triangle*>(enclosedObjects[k]);
    if(tri) {
      myfile << "mesh vertices:#(";
      myfile << "[" <<  tri->a.x << "," << tri->a.y << "," << tri->a.z << "],";
      myfile << "[" <<  tri->b.x << "," << tri->b.y << "," << tri->b.z << "],";
      myfile << "[" <<  tri->c.x << "," << tri->c.y << "," << tri->c.z << "]";
      myfile << ") faces:#([1,2,3]) material:meditMaterials[1]" << endl;
    }
  }

  if(true) {
    myfile << "a = a+1;" << endl;
    myfile << "aStr = formattedPrint a format: \"8.8d\"" << endl;
    myfile << "filename = file + aStr + ext;" << endl;
    myfile << "render outputfile:filename" << endl;
  } else {
    myfile << "messagebox \"this is level " << depth << " \" " << endl;
  }
  myfile.close();
}