// Copyright (C) Thorsten Thormaehlen

#ifndef DEF_BVHIERARCHY_H
#define DEF_BVHIERARCHY_H

#include <vector>
#include <algorithm>  
#include "SceneObjects.h"


  /*!
  \class BVHierarchyNode BVHierarchy.h
  \brief This class implements a Bounding Volume Hierarchy (BVH)
  
  The class BVHierarchyNode represents a node in a binary BVH tree. 
  Each node has pointers to its two children (or NULL-pointers for leaf nodes).
  Leaf nodes hold pointers to enclosed intersectable scene objects. 

  The BVH-tree is an intersectable scene object and acts as a replacement
  for all intersectable scene objects that are passed to the constructor.

  The BVH-tree takes full control over the passed intersectable scene objects, i.e., it
  stores their pointers and deletes the instances when the tree is deleted.
  */


  class BVHierarchyNode : public SceneObject, public Intersectable
  {    
  private:
    BoundingBox box;
    BVHierarchyNode *leftNode;
    BVHierarchyNode *rightNode; 
    std::vector <Intersectable*> enclosedObjects;

  public :
    //! constructor for an empty child node
    BVHierarchyNode() {
      leftNode = NULL;
      rightNode = NULL;
    } 

    //! constructor for a tree node (builds a BVH sub tree)
    BVHierarchyNode(std::vector<SceneObject*> &objects) {
      leftNode = NULL;
      rightNode = NULL;
      this->build(objects);
    } 

    //! destructor
    ~BVHierarchyNode() {
      if(leftNode) delete leftNode;
      if(rightNode) delete rightNode;
      for(unsigned i=0; i < enclosedObjects.size(); i++) {
        delete enclosedObjects[i];
      }
    } 

  public:

    //! find closest intersection 
    bool intersect(const Ray& ray, RayHit &hit) const {
      float closestDist = FLT_MAX;
      return depthSearch(ray, hit, closestDist);
    }

	bool intersect(const BoundingBox& other) const{
		return box.intersect(other);
	}

    //! perform a deep search in the tree to find the closest intersection
    bool depthSearch(const Ray& ray, RayHit &hit, float &closest) const;

	// perfroms a deeps earch in the tree to find the closest point to a given point 'p' within a boundingbox 'searchBox'
	bool getClosestPointInBoxDepthSearch(Vec3f& p, BoundingBox& searchBox, float& closestDistance, Vec3f& closestPoint) const;

	Triangle* getAdjacentEdgeDepthSearch(Triangle* excludeTriangle, BoundingBox& searchBox, Vec3f& edgeA, Vec3f& edgeB, float& closestDistance) const;

	bool getAdjacentEdge(Vec3f& a, Vec3f& b, Triangle* parentTriangle,Triangle*& resultParentTriangle) const;

	Vec3f getClosestPoint(const Vec3f &p) const;

	Vec3f getCenter() const;

	float getSurfaceArea() const;

    // returns the bounding box of the node
    BoundingBox getBoundingBox() const {
      return box;
    }

  private:

    // build a BVH tree for the given objects
    void build(std::vector<SceneObject*> &objects);

    // returns the bounding box around all given objects
    BoundingBox getUnionOfBoundingBoxes(std::vector<Intersectable*> &objects);

    // compute SAH according to Wald et al.
    float surfaceAreaHeuristic(unsigned sizeLeft, float leftArea, unsigned sizeRight, float rightArea);

    // build BVH according to Wald et al.
    void partitionSweep(std::vector<Intersectable*> &objects, unsigned depth);

    // for debugging purpose
    void debug(unsigned depth) const;
  };
#endif
