// Copyright (C) Thorsten Thormaehlen

#ifndef DEF_BOUNDINGBOX_H
#define DEF_BOUNDINGBOX_H

#include "Vec3.h"

  /*!
  \class BoundingBox BoundingBox.h
  \brief This class describes a bounding box
  */
  class BoundingBox
  {    
  public:
    Vec3f min;
    Vec3f max;

  public :
    //! constructor
    BoundingBox() {
      min = Vec3f(0.0,0.0,0.0);
      max = Vec3f(0.0,0.0,0.0);
    } 

    //! constructor
    BoundingBox(const Vec3f& minBoundingBoxVert, const Vec3f&  maxBoundingBoxVert) { 
      min = minBoundingBoxVert;
      max = maxBoundingBoxVert;
    } 

    bool intersect(const Ray& ray) const {
      const float eps = 1e-4f;
      Vec3f d = ray.direction;
      Vec3f p = ray.origin;

      float tmin, tmax, tymin, tymax, tzmin, tzmax;
      // handle x-direction
      if (ray.direction.x >= 0) {
        tmin = (min.x - p.x) / d.x;
        tmax = (max.x - p.x) / d.x;
      } else {
        tmin = (max.x - p.x) / d.x;
        tmax = (min.x - p.x) / d.x;
      }

      // handle y-direction
      if (ray.direction.y >= 0) {
        tymin = (min.y - p.y) / d.y;
        tymax = (max.y - p.y) / d.y;
      } else {
        tymin = (max.y - p.y) / d.y;
        tymax = (min.y - p.y) / d.y;
      }

      // check order
      if ( (tmin > tymax) || (tymin > tmax) ) return false;

      // find the minium of x and y
      if (tymin > tmin)
        tmin = tymin;
      if (tymax < tmax)
        tmax = tymax;

      // handle z-direction
      if (ray.direction.z >= 0) {
        tzmin = (min.z - p.z) / d.z;
        tzmax = (max.z - p.z) / d.z;
      } else {
        tzmin = (max.z - p.z) / d.z;
        tzmax = (min.z - p.z) / d.z;
      }

      // check order again
      if ( (tmin > tzmax) || (tzmin > tmax) ) return false;

      // find the minium of x, y, and y
      if (tzmin > tmin)
        tmin = tzmin;
      if (tzmax < tmax)
        tmax = tzmax;

      return (tmax > eps);
    }

	bool intersect(const BoundingBox& other) const{
		float maxOverlap = intervallDistance(min.x, max.x, other.min.x, other.max.x);
		float overlap = intervallDistance(min.y, max.y, other.min.y, other.max.y);
		if (overlap > maxOverlap)
			maxOverlap = overlap;
		overlap = intervallDistance(min.z, max.z, other.min.z, other.max.z);
		if (overlap > maxOverlap)
			maxOverlap = overlap;
		if (maxOverlap < 0)
			return true;
		return false;
	}

    //! computes the surface area
    float surfaceArea() const {
      Vec3f d = max - min;
      return 2.0f * (d.x * d.y + d.x * d.z + d.y * d.z);
    }

    //! returns a bounding box containing the box itself and the given other box
    BoundingBox getUnion(const BoundingBox &other) const {
      Vec3f newMin, newMax;
      newMin.x = (min.x < other.min.x)? min.x : other.min.x;
      newMin.y = (min.y < other.min.y)? min.y : other.min.y;
      newMin.z = (min.z < other.min.z)? min.z : other.min.z;
      newMax.x = (max.x > other.max.x)? max.x : other.max.x;
      newMax.y = (max.y > other.max.y)? max.y : other.max.y;
      newMax.z = (max.z > other.max.z)? max.z : other.max.z;
      return BoundingBox(newMin, newMax);
    }

    Vec3f center() const {
      return (max+min) * 0.5;
    }

	Vec3f closestPointInsideBox(Vec3f& p) {
		Vec3f result = p;
		if (result.x < min.x)
			result.x = min.x;
		if (result.y < min.y)
			result.y = min.y;
		if (result.z < min.z)
			result.z = min.z;

		if (result.x > max.x)
			result.x = max.x;
		if (result.y > max.y)
			result.y = max.y;
		if (result.z > max.z)
			result.z = max.z;

		return result;
	}

  private:
	  // a0--------b0       a1---------b1       -> no overlap -> returns positive value
	  // 
	  // a0--------b0
	  //      a1---------b1                      -> overlap -> returns negative value
	  float intervallDistance(float a0, float b0, float a1, float b1) const{
		  if (a0 < a1)
			  return a1 - b0;
		  else
			  return a0 - b1;
	  }
  };
#endif
