
#ifndef DEF_PROJECTIONRASTERIZER_H
#define DEF_PROJECTIONRASTERIZER_H

#include "Model.h"
#include "Vec3.h"
#include "SceneObjects.h"
#include <set>
#include <string>

using namespace std;

class ProjectionPath {
private:
	unsigned int* raster;
	unsigned int* edge;
	unsigned int resX, resY;
	Vec3f projectionU, projectionV;

	unsigned int rasterize(Model& model, Vec3f& projectionVector) {
		//span coordinate system
		Vec3f projection = projectionVector;
		projection.normalize();
		projectionU = projection.cross(Vec3f(1, 0, 0));
		if (projectionU.sqrLength() < 0.01)
			projectionU = projection.cross(Vec3f(0, 1, 0));
		if (projectionU.sqrLength() < 0.01)
			projectionU = projection.cross(Vec3f(0, 0, 1));
		if(projectionU.sqrLength()<0.01)
			projectionU = projection.cross(Vec3f(rand(), rand(), rand()));
		projectionU.normalize();
		projectionV = projection.cross(projectionU);
		projectionV.normalize();

		BoundingBox b = model.getBoundingBox();

		Ray r(Vec3f(0, 0, 0), projectionVector);
		r.direction.normalize();

		RayHit hit;

		unsigned int pixels = 0;

		//we assume that the model is noramlize to the range ([-0.5..0.5],[-0.5..0.5],[-0.5..0.5])
		for (unsigned int y = 0; y < resY; ++y) {
			for (unsigned int x = 0; x < resX; ++x) {
				r.origin = projectionU * (1.5f*x / resX - 0.75f) + projectionV * (1.5f*y / resY - 0.75f) - projection * 2.0; //world-space to projection-space
				if (model.getIntersection(r, hit)) {
					pixels++;
					raster[y*resX + x] = 1;
				}
				else {
					raster[y*resX + x] = 0;
				}
			}
		}
		return pixels;
	}

	//simple 4 neighbour edge detection filter
	unsigned int runEdgeDetection() {
		unsigned int pixels = 0;
		for (unsigned int y = 1; y < resY - 1; ++y) {
			for (unsigned int x = 1; x < resX - 1; ++x) {
				edge[y*resX + x] = 0;
				if (raster[y*resX + x]) {
					if (!(raster[(y - 1)*resX + x] && raster[(y + 1)*resX + x] && raster[y*resX + x + 1] && raster[y*resX + x - 1])) {
						edge[y*resX + x] = raster[y*resX + x];
						pixels++;
					}
				}
			}
		}
		return pixels;
	}

	//all directions in the form [x0,y0,x1,y1...] going clockwise around a 3x3 grid with (0,0) beeing the center.
	int directions[16] = { -1,-1,0,-1,1,-1,1,0,1,1,0,1,-1,1,-1,0 };
	bool getNextCoursorPos(unsigned int currentX, unsigned int currentY, unsigned int previousX, unsigned int previousY, unsigned int& nextX, unsigned int& nextY) {

		unsigned int startingIndex = 0;
		//getting the starting index from the previous position
		for (unsigned int i = 0; i < 8; ++i) {
			if (directions[i * 2] == previousX - currentX && directions[i * 2 + 1] == previousY - currentY) {
				startingIndex = i + 1;
				break;
			}
		}

		//going clockwise around the center from the starting position to find the next step
		unsigned int index;
		for (unsigned int i = 0; i < 8; ++i) {
			index = ((startingIndex + i) % 8) * 2;
			if (edge[(currentY + directions[index + 1]) * resX + currentX + directions[index]]) {
				nextX = currentX + directions[index];
				nextY = currentY + directions[index + 1];
				return true;
			}
		}

		return false;
	}

	bool travelPath(unsigned int startX, unsigned int startY, unsigned int segmentId) {
		unsigned int coursorX = startX;
		unsigned int coursorY = startY;

		unsigned int firstCoursorX = coursorX;
		unsigned int firstCoursorY = coursorY;
		unsigned int nextCoursorX = coursorX;
		unsigned int nextCoursorY = coursorY;
		unsigned int previousCoursorX = coursorX-1;
		unsigned int previousCoursorY = coursorY;

		vector<Vec3f> path;

		//projection-space to world-space
		segmentIds.push_back(segmentId);
		path.push_back(projectionU*(coursorX*(1.5f / resX) - 0.75f) + projectionV * (coursorY*(1.5f / resY) - 0.75f));

		bool wasAwayFromStartingPoint = false;

		while (getNextCoursorPos(coursorX, coursorY, previousCoursorX, previousCoursorY, nextCoursorX, nextCoursorY)) {
			previousCoursorX = coursorX;
			previousCoursorY = coursorY;
			coursorX = nextCoursorX;
			coursorY = nextCoursorY;


			segmentIds.push_back(segmentId);
			path.push_back(projectionU*(coursorX*(1.5f / resX) - 0.75f) + projectionV * (coursorY*(1.5f / resY) - 0.75f));

			if (path.size() > 32768) // something went wrong
				return false;

			if (abs((int)coursorX - (int)startX) < 2 && abs((int)coursorY - (int)startY) < 2) 
			{
				if (wasAwayFromStartingPoint) {

					if (path.size() > 64) { //ignore very small paths
						for (int i = 0; i < path.size(); ++i)
							this->path.push_back(path[i]);
					}
					return true; //true if nothing went wrong
				}
			}
			else {
				wasAwayFromStartingPoint = true;
			}

		}
		return false;
	}

	void floodFill(unsigned int* pixels, unsigned int pixelsWidth, int x, int y, int value) {
		vector<pair<unsigned int, unsigned int>>* activePixels = new vector<pair<unsigned int, unsigned int>>();
		vector<pair<unsigned int, unsigned int>>* neighbourPixels = new vector<pair<unsigned int, unsigned int>>();
		vector<pair<unsigned int, unsigned int>>* h;

		activePixels->push_back(pair<unsigned int, unsigned int>(x, y));

		unsigned int fillValue = pixels[y*pixelsWidth + x];

		while (activePixels->size() > 0) {
			neighbourPixels->clear();
			for (int i = 0; i < activePixels->size(); ++i) {
				int px = (*activePixels)[i].first;
				int py = (*activePixels)[i].second;
				pixels[py*pixelsWidth + px] = value;
				for (int xi = px-1; xi <= px+1; ++xi) {
					for (int yi = py-1; yi <= py+1; ++yi) {
						if (pixels[yi*pixelsWidth + xi] == fillValue) {
							neighbourPixels->push_back(pair<unsigned int, unsigned int>(xi, yi));
						}
					}
				}
			}
			h = activePixels;
			activePixels = neighbourPixels;
			neighbourPixels = h;
		}
	}

	//creates a path our of the edge array (rasterization and edge detection has to be done first)
	bool createOutlinePath() {
		path.clear();
		segmentIds.clear();

		//find starting point on edge of the rasterized object
		unsigned int segments = 0;
		for (unsigned int y = 1; y < resY - 1; ++y) {
			for (unsigned int x = 1; x < resX - 1; ++x) {
				if (edge[y*resX + x]) {
					segments++;
					if(!travelPath(x, y,segments))
						return false;
					floodFill(edge, resX, x, y,0);
				}
			}
		}
		return segments>0;
	}

	//returns a value in [0..1], where 1 is maximum spiky and 0 ist a straight line
	//the neighbourhood parameter determines how many points away from index are looked at to determine the average direction the line segment is going. 
	float getCurvature(unsigned int index, unsigned int neighbourhood = 8) {
		Vec3f p = path[index];
		unsigned int idx;
		unsigned int reverseWeight = 1;

		Vec3f leftDirection;
		for (int i = index - 1; i > index - neighbourhood - 1; --i) {
			idx = (i + path.size()) % path.size();
			leftDirection = leftDirection + (path[idx] - p)*(1.0f/reverseWeight);
			reverseWeight++;
		}
		leftDirection.normalize();


		Vec3f rightDirection;
		reverseWeight = 1;
		for (int i = index + 1; i < index + neighbourhood + 1; ++i) {
			idx = i % path.size();
			rightDirection = rightDirection + (path[idx] - p)*(1.0f / reverseWeight);
			reverseWeight++;
		}
		rightDirection.normalize();

		return (rightDirection.dot(leftDirection) + 1.0)*0.5f;
	}

	void calculateCurvature() {
		//calculate curvature on each point
		curvature.clear();
		for (int i = 0; i < path.size(); ++i) {
			curvature.push_back(getCurvature(i));
		}

		//finding high curvature situations
		highCurvaturePoints.clear();
		int maxCourveIndex;
		float maxCourve;		
		for (int i = 0; i < path.size(); ++i) {
			maxCourveIndex = 0;
			maxCourve = 0;
			for (int di = -8; di < 8; ++di) {
				if (curvature[(i + di + path.size()) % path.size()] > maxCourve) {
					maxCourveIndex = (i + path.size() + di) % path.size();
					maxCourve = curvature[maxCourveIndex];
				}
			}
			if (i == maxCourveIndex && maxCourve > 0.18)
				highCurvaturePoints.push_back(i);
		}
	}

public:
	
	vector<Vec3f> path;
	vector<float> curvature;
	vector<unsigned int> segmentIds;

	vector<int> highCurvaturePoints;

	ProjectionPath(unsigned int resolutionX, unsigned int resolutionY) {
		resX = resolutionX;
		resY = resolutionY;
		raster = new unsigned int[resolutionX*resolutionY];
		edge = new unsigned int[resolutionX*resolutionY];
	}

	~ProjectionPath() {
		delete[] raster;
		delete[] edge;
	}

	bool buildProjectionPath(Model& model, Vec3f& projectionVector) {
		unsigned int raster = rasterize(model, projectionVector);
		unsigned int edge = runEdgeDetection();
		bool success = createOutlinePath();
		calculateCurvature();
		return success;
	}

};

#endif