
#ifndef DEF_PROJECTIONPATH_H
#define DEF_PROJECTIONPATH_H

#include "Model.h"
#include "Vec3.h"
#include "SceneObjects.h"
#include <set>
#include <string>

using namespace std;

class Edge {

public:
	Vec3f a, b;
	Edge(Vec3f& _a,Vec3f& _b) {
		a = _a;
		b = _b;
	}

	//! < operator
	bool operator<(const Edge &op) const
	{
		Vec3f minThis = min(a, b);
		Vec3f maxThis = max(a, b);
		Vec3f minOp = min(op.a, op.b);
		Vec3f maxOp = max(op.a, op.b);
		if (minThis < minOp || minOp < minThis)
			return minThis < minOp;
		return maxThis<maxOp;
	}
};

class ProjectionPathLegacy {

private:
	Vec3f projectionVector;
	set<Edge> silhouette;
	vector<Vec3f> path;
	const Model* model;

	//returns the vertex position of the triangle t that is not on edge edgeA-edgeB
	Vec3f getOpposingVertex(Vec3f& edgeA, Vec3f& edgeB, Triangle* t) {
		Vec3f opposingVertex = t->a;
		float distanceA = (opposingVertex - edgeA).sqrLength();
		float distanceB = (opposingVertex - edgeB).sqrLength();
		float minAB = min(distanceA, distanceB);
		float maxDistance = minAB;

		distanceA = (t->b - edgeA).sqrLength();
		distanceB = (t->b - edgeB).sqrLength();
		minAB = min(distanceA, distanceB);

		if (minAB > maxDistance) {
			maxDistance = minAB;
			opposingVertex = t->b;
		}

		distanceA = (t->c - edgeA).sqrLength();
		distanceB = (t->c - edgeB).sqrLength();
		minAB = min(distanceA, distanceB);

		if (minAB > maxDistance) {
			maxDistance = minAB;
			opposingVertex = t->c;
		}

		return opposingVertex;
	}

	void processEdge(Vec3f& a, Vec3f& b, Triangle* t) {
		
		Triangle* other;

		if (!model->getAdjacentEdge(a, b, t, other))
			return;

		//find the opposing vertices va, vb
		Vec3f va = getOpposingVertex(a, b, t);
		Vec3f vb = getOpposingVertex(a, b, other);

		//test the edge against the projection vector to see if it is part of the silhouette
		Vec3f n1 = (b - a).cross(va - a);
		Vec3f n2 = (vb - a).cross(b - a);
		float d1 = n1.dot(projectionVector);
		float d2 = n2.dot(projectionVector);
		if (d1 <=0 &&d2>0 || d1> 0 && d2<=0) {
			//the edge a-b is part of the silhouette
		}
		else {
			//the edge a-b is not part of the silhouette or the second triangle containing this edge has not been processed, yet.
			return;
		}

		Vec3f projectedA = a - projectionVector * a.dot(projectionVector);
		Vec3f projectedB = b - projectionVector * b.dot(projectionVector);
		if (projectedA == projectedB)
			return;

		// if part of the siluette, project the edge with the projection vector (orthographic) and save it as part of the projection result
		silhouette.insert(Edge(projectedA, projectedB));
	}

	void buildSillhouette(const Model& m) {
		silhouette.clear();

		Triangle* t;
		for (int i = 0; i < m.triangles.size(); ++i) {
			t = m.triangles[i];

			processEdge(t->a, t->b, t);
			processEdge(t->a, t->c, t);
			processEdge(t->b, t->c, t);
		}
	}

	bool pathDepthSearch(map<Vec3f, vector<Vec3f>>& graph, const Vec3f& previousPosition, const Vec3f& currentPosition, const Vec3f& startingPoision, vector<Vec3f>& path) {
		for (int i = 0; i < graph[currentPosition].size(); ++i) {
			if (graph[currentPosition][i] == previousPosition)
				continue;
			if (graph[currentPosition][i] == startingPoision) {
				path.push_back(currentPosition);
				return true;
			}
			else if(pathDepthSearch(graph,currentPosition, graph[currentPosition][i],startingPoision,path)){
				path.push_back(currentPosition);
				return true;
			}
		}
		return false;
	}

	void constructPath() {
		path.clear();

		//build graph
		map<Vec3f, vector<Vec3f>> graph;
		for (auto edge : silhouette) {
			if (graph.count(edge.a) <= 0)
				graph[edge.a] = vector<Vec3f>();
			if (graph.count(edge.b) <= 0)
				graph[edge.b] = vector<Vec3f>();

			graph[edge.a].push_back(edge.b);
			graph[edge.b].push_back(edge.a);
		}

		//deep search the graph for a circle from a random start point to itself
		pathDepthSearch(graph, silhouette.begin()->b, silhouette.begin()->a, silhouette.begin()->a, path);
	}

public:
	ProjectionPathLegacy() {

	}

	ProjectionPathLegacy(const Model& m, const Vec3f& _projectionVector) {
		build(m, _projectionVector);
	}

	void build(const Model& m, const Vec3f& _projectionVector) {
		model = &m;
		projectionVector = _projectionVector;
		projectionVector.normalize();
		buildSillhouette(m);
		constructPath();
	}

	string toString() {
		string s = "";
		for (int i = 0; i < path.size(); ++i) {
			s += "("+ to_string(path[i].x) +"/"+ to_string(path[i].y)+"/"+ to_string(path[i].z)+")";
		}
		return s;
	}
};

#endif