// Copyright (C) Thorsten Thormaehlen
// Modified and extended by Sebastian Lieb

#ifndef DEF_SCENEOBJECTS_H
#define DEF_SCENEOBJECTS_H

#define _USE_MATH_DEFINES
#include <math.h>
#include <algorithm>

#include "Vec3.h"
#include "Ray.h"
#include "BoundingBox.h"

  /*!
  \class SceneObject SceneObjects.h
  \brief This class is the base class for an object of a scene
  */

  class SceneObject
  {

  protected :
    Vec3f pos;

  public :
    //! constructor
    SceneObject(const Vec3f &p = Vec3f(0.0f, 0.0f, 0.0f))
    { 
      pos = p;
    }

    virtual Vec3f getPos() const
    {
      return pos;
    }
  };

  /*!
  \class Intersectable SceneObjects.h
  \brief This class is an abstract base class for an intersectable object
  */

  class Intersectable
  {
  protected :

  public :
    virtual bool intersect(const Ray& ray, RayHit &hit) const = 0;
	virtual bool intersect(const BoundingBox& BB) const = 0;
    virtual BoundingBox getBoundingBox() const = 0;
	virtual Vec3f getClosestPoint(const Vec3f &p) const = 0;
	virtual Vec3f getCenter() const = 0;
	virtual float getSurfaceArea() const = 0;
  };


  /*!
  \class Triangle SceneObjects.h
  \brief This class describes a triangle
  */
  class Triangle : public SceneObject, public Intersectable
  {    
  public:
    Vec3f a, b, c; // triangle vertices

  public :
    //! constructor
    Triangle() { 
    } 

    //! constructor
    Triangle(const Vec3f& p1, const Vec3f& p2, const Vec3f &p3):Triangle() { 
      a = p1;
      b = p2;
      c = p3;
      // compute normal from cross product
    } 


	//get the closest point on the triangle to a given point p
	Vec3f getClosestPoint(const Vec3f &p) const
	{
		Vec3f edge0 = b - a;
		Vec3f edge1 = c - a;
		Vec3f v0 = a - p;

		float af = edge0.dot(edge0);
		float bf = edge0.dot(edge1);
		float cf = edge1.dot(edge1);
		float df = edge0.dot(v0);
		float ef = edge1.dot(v0);

		float det = af * cf - bf * bf;
		float s = bf * ef - cf * df;
		float t = bf * df - af * ef;

		if (s + t < det)
		{
			if (s < 0.f)
			{
				if (t < 0.f)
				{
					if (df < 0.f)
					{
						s = clamp01(-df / af);
						t = 0.f;
					}
					else
					{
						s = 0.f;
						t = clamp01(-ef / cf);
					}
				}
				else
				{
					s = 0.f;
					t = clamp01(-ef / cf);
				}
			}
			else if (t < 0.f)
			{
				s = clamp01(-df / af);
				t = 0.f;
			}
			else
			{
				float invDet = 1.f / det;
				s *= invDet;
				t *= invDet;
			}
		}
		else
		{
			if (s < 0.f)
			{
				float tmp0 = bf + df;
				float tmp1 = cf + ef;
				if (tmp1 > tmp0)
				{
					float numer = tmp1 - tmp0;
					float denom = af - 2 * bf + cf;
					s = clamp01(numer / denom);
					t = 1 - s;
				}
				else
				{
					t = clamp01(-ef / cf);
					s = 0.f;
				}
			}
			else if (t < 0.f)
			{
				if (af + df > bf + ef)
				{
					float numer = cf + ef - bf - df;
					float denom = af - 2 * bf + cf;
					s = clamp01(numer / denom);
					t = 1 - s;
				}
				else
				{
					s = clamp01(-ef / cf);
					t = 0.f;
				}
			}
			else
			{
				float numer = cf + ef - bf - df;
				float denom = af - 2 * bf + cf;
				s = clamp01(numer / denom);
				t = 1.f - s;
			}
		}

		return a + edge0 * s + edge1 * t;
	}

    bool intersect(const Ray& ray, RayHit &hit) const {
	  const float eps = 0.0f;// 1e-2f;

      // here we use exactly the notation from the lecture slides
      // a, b, and c are the three triangle vertices
      Vec3f d = ray.direction;
      Vec3f p = ray.origin;

      Vec3f u = b - a;
      Vec3f v = c - a;
      Vec3f w = p - a;

      // implementing ray/triangle intersection according to
      // the lecture slides by computing
      // (r, s, t) = (1 / (d x v) * u) ((w x u) * v, (d x v) *w, (w x u) * d)

      Vec3f dxv = d.cross(v);
      float dxvu = dxv.dot(u);

      // denominator close to zero?
      if (fabs(dxvu) < 1e-16) return false;
      float f = 1.0f / dxvu;

      // compute s
      float s = dxv.dot(w) * f;
      if (s < 0.0f || s > 1.0f) return false;

      // compute t
      Vec3f wxu = w.cross(u);
      float t =  wxu.dot(d) * f;
      if (t < 0.0 || t > 1.0) return false;
      if(s + t > 1.0) return false;

      // compute r
      float r =  wxu.dot(v) * f;
      if (r < eps) return false;

      hit.hitPoint = p + d * r; 
      return true;
    }

	bool intersect(const BoundingBox& BB) const{
		return getBoundingBox().intersect(BB);
	}

        
    BoundingBox getBoundingBox() const {
      Vec3f min, max;
      min.x = (a.x < b.x)? a.x : b.x;
      min.x = (min.x < c.x)? min.x : c.x;
      min.y = (a.y < b.y)? a.y : b.y;
      min.y = (min.y < c.y)? min.y : c.y;
      min.z = (a.z < b.z)? a.z : b.z;
      min.z = (min.z < c.z)? min.z : c.z;
      max.x = (a.x > b.x)? a.x : b.x;
      max.x = (max.x > c.x)? max.x : c.x;
      max.y = (a.y > b.y)? a.y : b.y;
      max.y = (max.y > c.y)? max.y : c.y;
      max.z = (a.z > b.z)? a.z : b.z;
      max.z = (max.z > c.z)? max.z : c.z;
      return BoundingBox(min, max);
    }

	Vec3f getCenter() const{
		return (a + b + c)*0.3333f;
	}

	float getSurfaceArea() const{
		return (b - a).cross(c - a).length()*0.5f;
	}

	private:
		float clamp01(float value) const{
			return std::max(0.0f, std::min(1.0f, value));
		}
  };
#endif
