// Copyright (C) Thorsten Thormaehlen

#ifndef DEF_VEC3_H
#define DEF_VEC3_H
#include <math.h> 

#define equivalentEpsilon 0.001f

  /*!
  \class Vec3 Vec3.h
  \brief This class defines a templated 3-vector.
  */

  template<typename T> class Vec3
  {
  public :

    union {
      struct { // human-friendly access
        T x;
        T y;
        T z;
      };
      T xyz[3];
      struct { // human-friendly access
        T r; // red
        T g; // green
        T b; // blue
      };
      T rgb[3];
    };

  public :
    //! constructor
    Vec3(T xx=0, T yy=0, T zz=0)
    { 
      x= xx; y = yy; z = zz; 
    } 

    //! destructor
    ~Vec3() {}

    //! plus operator
    Vec3<T> operator+(const Vec3<T> &op) const 
    { 
      return Vec3<T>(x+op.x, y+op.y, z+op.z); 
    } 

	//! minus operator
	Vec3<T> operator-(const Vec3<T> &op) const
	{
		return Vec3<T>(x - op.x, y - op.y, z - op.z);
	}

	//! < operator
	bool operator<(const Vec3<T> &op) const
	{
		if (fabsf(x - op.x) > equivalentEpsilon)
			return x < op.x;
		if (fabsf(y - op.y) > equivalentEpsilon)
			return y < op.y;
		if (fabsf(z - op.z) > equivalentEpsilon)
			return z < op.z;
		return false;
	}

	//! == operator
	bool operator==(const Vec3<T> &op) const
	{
		return fabsf(x - op.x)< equivalentEpsilon && fabsf(y - op.y) < equivalentEpsilon && fabsf(z - op.z) < equivalentEpsilon;
	}

    //! scalar multiplication
    Vec3<T> operator*(T s) const 
    { 
      return Vec3<T>(x*s, y*s, z*s); 
    }

    //! vector component-wise multiplication
    Vec3<T> operator*(const Vec3<T> &op) const 
    { 
      return Vec3<T>(x*op.x, y*op.y, z*op.z); 
    } 

    //! return length of vector
    T length() const 
    { 
      return sqrt(x*x+y*y+z*z);
    } 

	//! return length*length of vector
	T sqrLength() const
	{
		return x*x + y * y + z * z;
	}

	//! sets all components of the vector to n
	void set(T n) {
		x = n; y = n; z = n;
	}

    //! normalize the vector
    void normalize() 
    { 
      T n = T(1.0)/length();
      x = n * x;  y = n * y;  z = n * z; 
    } 

    //! scalar product
    T dot(const Vec3<T> &op) const 
    { 
      return (x*op.x + y*op.y + z*op.z); 
    } 

    //! cross product
    Vec3<T> cross(const Vec3<T> &op) const 
    { 
      return Vec3<T>(y * op.z - z * op.y,
        z * op.x - x * op.z,
        x * op.y - y * op.x);
    } 

    //! clamp vector to low and high value 
    void clamp(T low, T high) 
    { 
      if(x < low) x = low;
      if(x > high) x = high;
      if(y < low) y = low;
      if(y > high) y = high;
      if(z < low) z = low;
      if(z > high) z = high;
    } 

	int getHash() {
		return *((unsigned int)(x*32)) + *((unsigned int)(y*32) << 5) + *((unsigned int)(z*32) << 11);
	}
  };

  typedef Vec3<float> Vec3f;
  typedef Vec3<double> Vec3d;
  typedef Vec3<int> Vec3i;

#endif
